#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#   Copyright 2017 Kaede Hoshikawa
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import Optional, Callable, Any

import os
import asyncio
import functools


class TestHelper:
    def __init__(self, file: str) -> None:
        self._root_path = os.path.dirname(os.path.abspath(file))

        self._loop: Optional[asyncio.AbstractEventLoop] = None

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        if self._loop is None:
            self._loop = asyncio.get_event_loop()

            self._loop.set_debug(True)

        return self._loop

    def abspath(self, sub_path: str) -> str:
        return os.path.abspath(
            os.path.realpath(
                os.path.join(self._root_path, sub_path)))

    def force_sync(self, fn: Callable[..., Any]) -> Callable[..., Any]:
        @functools.wraps(fn)
        def wrapper(_self: Any, *args: Any, **kwargs: Any) -> Any:
            return self.loop.run_until_complete(
                asyncio.wait_for(
                    fn(_self, *args, **kwargs), 10))

            # Wait 10 sec, or kill the task.

        return wrapper



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#   Copyright 2017 Kaede Hoshikawa
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import AsyncIterator

from sketchbook import Sketch
from sketchbook.testutils import TestHelper

helper = TestHelper(__file__)

import time
import random
import asyncio
import pytest


class _AsyncTimeIterator(AsyncIterator[float]):
    def __init__(self) -> None:
        self._time = time.time()

        self._counter_left = random.choice(range(2, 5))

        self._iterated_time = []

    def __aiter__(self) -> "_AsyncTimeIterator":
        return self

    async def __anext__(self) -> float:
        await asyncio.sleep(random.choice(range(1, 5)) / 100)

        if self._counter_left > 0:
            self._counter_left -= 1

            current_time = time.time()
            self._iterated_time.append(current_time)

            return current_time

        else:
            raise StopAsyncIteration


class IfElifElseTestCase:
    @helper.force_sync
    async def test_if_elif_else(self) -> None:
        skt = Sketch(
            "<% if cond %>cond_str<% elif sub_cond %>sub_cond_str"
            "<% else %>else_str<% end %>")

        assert await skt.draw(cond=True, sub_cond=True) == "cond_str"

        assert await skt.draw(cond=False, sub_cond=True) == "sub_cond_str"

        assert await skt.draw(cond=False, sub_cond=False) == "else_str"


class AsyncForTestCase:
    @helper.force_sync
    async def test_async_for(self) -> None:
        time_iter = _AsyncTimeIterator()
        skt = Sketch(
            "<% async for i in time_iter %><%r= str(i) %>, <% end %>")

        assert await skt.draw(time_iter=time_iter) == \
            str(time_iter._iterated_time)[1:-1] + ", "


class RaiseErrorTestCase:
    @helper.force_sync
    async def test_raise_error(self):
        skt = Sketch("<% raise RuntimeError %>")
        with pytest.raises(RuntimeError):
            await skt.draw()

class VariableAssignmentTestCase:
    @helper.force_sync
    async def test_assign_var(self):
        skt = Sketch("<% let a = b %><%= a %>")

        assert await skt.draw(b="I am b!") == "I am b!"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#   Copyright 2017 Kaede Hoshikawa
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from sketchbook import Sketch, SketchContext
from sketchbook.testutils import TestHelper

helper = TestHelper(__file__)


class OutputTestCase:
    @helper.force_sync
    async def test_html_escape(self) -> None:
        skt = Sketch("Hello, <%html= a %>!")

        assert await skt.draw(a="<h1>world</h1>") == \
            "Hello, &lt;h1&gt;world&lt;/h1&gt;!"

    @helper.force_sync
    async def test_no_escape(self) -> None:
        skt = Sketch("Hello, <%r= a %>!")

        assert await skt.draw(a="<h1>world</h1>") == "Hello, <h1>world</h1>!"

    @helper.force_sync
    async def test_json_escape(self) -> None:
        skt = Sketch('{"name": <%json= name %>}')

        assert await skt.draw(name="{\"name\": \"admin\"}") == \
            '{\"name\": \"{\\"name\\": \\"admin\\"}\"}'

    @helper.force_sync
    async def test_url_escape(self) -> None:
        skt = Sketch("https://www.example.com/?user=<%u= name %>")
        assert await skt.draw(name="a&redirect=https://www.example2.com/") == \
            ("https://www.example.com/?user=a%26redirect%3Dhttps%3A%2F%2F"
             "www.example2.com%2F")

    @helper.force_sync
    async def test_url_without_plus_escape(self) -> None:
        skt = Sketch("https://www.example.com/?user=<%url_without_plus= name %>")
        assert await skt.draw(name="John Smith") == \
            "https://www.example.com/?user=John%20Smith"


    @helper.force_sync
    async def test_custom_escape_fn(self) -> None:
        def _number_fn(i: int) -> str:
            return str(i + 1)

        skt_ctx = SketchContext(custom_escape_fns={"number": _number_fn})
        skt = Sketch("The result is <% number= a %>.", skt_ctx=skt_ctx)

        assert await skt.draw(a=12345) == "The result is 12346."



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#   Copyright 2017 Kaede Hoshikawa
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from sketchbook import SketchFinder, SketchContext, \
    SketchNotFoundError
from sketchbook.testutils import TestHelper

import pytest

helper = TestHelper(__file__)


class SketchFinderTestCase:
    def test_init(self) -> None:
        finder = SketchFinder(helper.abspath("sketches"))

        assert finder._root_path == helper.abspath("sketches") + "/"

        with pytest.raises(AssertionError):
            SketchFinder(-1)

    @helper.force_sync
    async def test_find(self) -> None:
        loader = SketchFinder(helper.abspath("sketches"))

        with pytest.raises(SketchNotFoundError):
            await loader.find("phantasm.html")
        loaded_skt = await loader.find("index.html")

        with open(helper.abspath("sketches/index.html")) as f:
            skt_content = f.read()
        assert loaded_skt._content == skt_content  # Test Correctness.

        sec_loaded_skt = await loader.find("index.html")
        assert loaded_skt is sec_loaded_skt  # Test Sketch Cache.

    @helper.force_sync
    async def test_find_no_cache(self) -> None:
        skt_ctx = SketchContext(cache_sketches=False)
        loader = SketchFinder(helper.abspath("sketches"), skt_ctx=skt_ctx)

        loaded_skt = await loader.find("index.html")
        sec_loaded_skt = await loader.find("index.html")

        # Test they have the same content.
        assert loaded_skt._content == sec_loaded_skt._content

        # Test Sketch Cache Disabled.
        assert loaded_skt is not sec_loaded_skt


class InheritanceTestCase:
    @helper.force_sync
    async def test_inherit(self) -> None:
        finder = SketchFinder(helper.abspath("sketches"))

        skt = await finder.find("index.html")

        assert await skt.draw() == """\
<!DOCTYPE HTML>
<html>
    <head>
        <title>Index Title</title>
    </head>
    <body>
        \n
This is body. The old title is Old Title.

    </body>
</html>
"""


class IncludeTestCase:
    @helper.force_sync
    async def test_include(self) -> None:
        finder = SketchFinder(helper.abspath("sketches"))

        skt = await finder.find("main.html")

        assert await skt.draw() == """\
<!DOCTYPE HTML>
<html>
    <head>
        <title>Main Page</title>
    </head>

    <body>

        \n<nav>This will be included into other files.</nav>

    </body>
</html>
"""


class SketchDiscoveryTestCase:
    @helper.force_sync
    async def test_traversal_prevention(self) -> None:
        finder = SketchFinder(helper.abspath("sketches"))

        with pytest.raises(SketchNotFoundError):
            await finder._find_abs_path(
                "../hijack.html", helper.abspath("sketches/main.html"))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#   Copyright 2017 Kaede Hoshikawa
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import Any, Callable

import html
import urllib.parse
import functools
import json
import collections

__all__ = ["builtin_escape_fns"]


def _escape_html(unsafe_str: str) -> str:
    assert isinstance(unsafe_str, str), \
            (f"The content({unsafe_str!r}) subject for html "
             f"escaping is not string.")
    return html.escape(unsafe_str)


def _no_escape(unsafe_str: str) -> str:
    assert isinstance(unsafe_str, str), \
            f"The variable({unsafe_str!r}) subject for output is not string."
    return unsafe_str


def _escape_url_with_plus(unsafe_str: str) -> str:
    assert isinstance(unsafe_str, str), \
            (f"The content({unsafe_str!r}) subject for url "
             f"escaping is not string.")
    return urllib.parse.quote_plus(unsafe_str)


def _escape_url_without_plus(unsafe_str: str) -> str:
    assert isinstance(unsafe_str, str), \
            (f"The content({unsafe_str!r}) subject for url "
             f"escaping is not string.")
    return urllib.parse.quote(unsafe_str)


def _escape_json(unsafe_var: Any) -> str:
    return json.dumps(unsafe_var)


builtin_escape_fns: "collections.OrderedDict[str, Callable[[Any], str]]" = \
    collections.OrderedDict([
        ("default", _escape_html),
        ("html", _escape_html),
        ("h", _escape_html),
        ("raw", _no_escape),
        ("r", _no_escape),
        ("url_with_plus", _escape_url_with_plus),
        ("url", _escape_url_with_plus),
        ("u", _escape_url_with_plus),
        ("url_without_plus", _escape_url_without_plus),
        ("json", _escape_json),
        ("j", _escape_json)
    ])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#   Copyright 2017 Kaede Hoshikawa
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import Optional, Mapping, Callable, Sequence, Type, Any

from . import escaping
from . import statements

import asyncio
import types

__all__ = ["SketchContext"]


class SketchContext:
    """
    :class:`.SketchContext` is used to hold options for :class:`.Sketch` and
    :class:`.BaseSketchFinder`, in order to alter the default behaviours of
    them.

    This is class should be immutable after initialization.

    :arg cache_sketches: If :code:`True`, :class:`.BaseSketchFinder` will
        cache the sketches. Default: :code:`True`.
    :arg source_encoding: The encoding of the source of the sketches if passed
        as bytestring. Default: :code:`utf-8`.
    :arg custom_escape_fns: Mapping of custom escape functions. Functions in
        this mapping will override the one with the same name in the built-in
        escape functions. Default: :code:`{}`.
    :arg loop: The event loop used by :class:`.Sketch` and
        :class:`.BaseSketchFinder`, must be a subclass of
        :class:`asyncio.AbstractEventLoop` or :code:`None`.
        Default: :code:`None`
        (Use the value of :func:`asyncio.get_event_loop`).

    Built-in Escape Functions:

    - :code:`default`, :code:`h`, and :code:`html`:
        Short hand for :func:`html.escape`.
    - :code:`r` and :code:`raw`:
        Output the input with no modification.
    - :code:`j` and :code:`json`:
        Short hand for :func:`json.dumps`.
    - :code:`u`, :code:`url` and :code:`url_with_plus`:
        Short hand for :func:`urllib.parse.quote_plus`.
    - :code:`url_without_plus`:
        Short hand for :func:`urllib.parse.quote`.
    """
    def __init__(
        self, *, cache_sketches: bool=True,
        source_encoding: str="utf-8",
        custom_escape_fns: Mapping[str, Callable[[Any], str]]={},
            loop: Optional[asyncio.AbstractEventLoop]=None) -> None:

        self._loop = loop or asyncio.get_event_loop()

        self._source_encoding = source_encoding

        escape_fns = escaping.builtin_escape_fns.copy()
        if custom_escape_fns:
            escape_fns.update(custom_escape_fns)
        self._escape_fns = types.MappingProxyType(escape_fns)

        self._stmt_classes = list(statements.builtin_stmt_classes)

        class OutputStmt(statements.BaseOutput):
            _filter_fn_names = list(self.escape_fns.keys())

        self._stmt_classes.append(OutputStmt)

        self._cache_sketches = cache_sketches

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        return self._loop

    @property
    def source_encoding(self) -> str:
        return self._source_encoding

    @property
    def escape_fns(self) -> Mapping[str, Callable[[Any], str]]:
        return self._escape_fns

    @property
    def stmt_classes(self) -> Sequence[Type[statements.Statement]]:
        return self._stmt_classes

    @property
    def cache_sketches(self) -> bool:
        return self._cache_sketches



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#   Copyright 2017 Kaede Hoshikawa
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import Union, Optional, Mapping, Any
from types import CodeType

from . import context
from . import finders
from . import statements
from . import parser
from . import runtime
from . import printer


__all__ = ["Sketch"]


class Sketch:
    """
    A compiled, reusable template object.

    :class:`.Sketch` can be initialized directly with arguments, or initialized
    by a subclass of :class:`.BaseSketchFinder`.

    :arg __content: The content of this sketch. This can be a string or a
        bytestring. If a bytestring
        is passed, the content will be decoded with :code:`source_encoding`
        option from :class:`.SketchContext`. This argument must be passed
        positionally and must be the first argument.
    :arg path: The path of the sketch, used by :class:`.SketchFinder` to
        resolve the file relationship. Default: :code:`<string>`.
    :arg skt_ctx: The :class:`.SketchContext` to be used by the
        :class:`.Sketch` Default: :code:`None`
        (Create a new :class:`.SketchContext` upon initialization).
    :arg finder: The finder used by the current sketch to include or inherit
        from other sketches. Default: :code:`None`.
    """
    def __init__(
        self, __content: Union[str, bytes], *,
        path: str="<string>",
        skt_ctx: Optional["context.SketchContext"]=None,
            finder: Optional["finders.BaseSketchFinder"]=None) -> None:
        self._path = path

        self._ctx = skt_ctx or context.SketchContext()

        self._finder = finder

        if isinstance(__content, bytes):
            self._content = __content.decode(self._ctx.source_encoding)

        else:
            self._content = __content

        self._root = parser.SketchParser.parse_sketch(self)

    @property
    def _compiled_code(self) -> CodeType:
        if not hasattr(self, "_printed_skt"):
            self._printed_skt = printer.PythonPrinter.print_sketch(self)

        return self._printed_skt

    def _get_runtime(
            self, skt_globals: Mapping[str, Any]) -> runtime.SketchRuntime:
        skt_globals = skt_globals

        exec(self._compiled_code, skt_globals)  # type: ignore

        return skt_globals["_SktCurrentRuntime"](
            self, skt_globals=skt_globals)

    async def draw(self, **kwargs: Any) -> str:
        """
        Draw the sketch to :code:`str`.

        :arg \*\*kwargs: All the keyword arguments will become global variables
            in the runtime.

        .. warning::

            The exceptions raised in the runtime will pop up from this method.
        """
        runtime = self._get_runtime(skt_globals=kwargs)

        await runtime._draw()

        return runtime._skt_result



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#   Copyright 2017 Kaede Hoshikawa
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from sketchbook import Sketch, SketchSyntaxError, UnknownStatementError
from sketchbook.testutils import TestHelper

import pytest

helper = TestHelper(__file__)

class StatementEscapeTestCase:
    @helper.force_sync
    async def test_stmts_escape(self) -> None:
        skt = Sketch(
            "<%% is the begin mark, and <%r= \"%%> is the end mark. \" %>"
            "<%r= \"<% and\" %> %> only need to be escaped whenever they "
            "have ambiguity of the templating system.")

        assert await skt.draw() == (
            "<% is the begin mark, and %> is the end mark. "
            "<% and %> only need to be escaped whenever they "
            "have ambiguity of the templating system.")

    @helper.force_sync
    async def test_multiline_stmts(self) -> None:
        skt = Sketch("""\
<% if a == \
            True
%>
Hello, it's me!
<% else %>
No, it's not me!
<% end %>""")

        assert await skt.draw(a=True) == "\nHello, it's me!\n"
        assert await skt.draw(a=False) == "\nNo, it's not me!\n"


class MalformedSketchTestCase:
    def test_malformed_stmts(self) -> None:
        with pytest.raises(SketchSyntaxError):
            Sketch("<% <%")

    def test_missing_end_mark(self) -> None:
        with pytest.raises(SketchSyntaxError):
            Sketch("<% while True %>")

    def test_redundant_end_mark(self) -> None:
        with pytest.raises(SketchSyntaxError):
            Sketch("<% if False %><% end %><% end %>")

    def test_unknown_stmt(self) -> None:
        with pytest.raises(UnknownStatementError):
            Sketch("<% if anyways %><% fi %>")

    def test_bad_assignment(self) -> None:
        with pytest.raises(SketchSyntaxError):
            Sketch("<% let a = b = c = d %>")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#   Copyright 2017 Kaede Hoshikawa
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import Optional, Union, Dict

from . import context
from . import sketch
from . import exceptions

import asyncio
import abc
import os
import concurrent.futures
import aiofiles

__all__ = ["BaseSketchFinder", "SketchFinder"]


class BaseSketchFinder(abc.ABC):
    """
    Base Sketch Finder.

    To create a custom sketch finder, subclass this class and override
    corresponding methods.

    :arg skt_ctx: The :class:`.SketchContext` to be used by the
        :class:`.BaseSketchFinder` and :class:`.Sketch`. Default: :code:`None`
        (Create a new :class:`.SketchContext` upon initialization).
    """
    def __init__(
        self, *,
            skt_ctx: Optional["context.SketchContext"]=None) -> None:
        self._ctx = skt_ctx or context.SketchContext()
        self._skt_cache: Dict[str, sketch.Sketch] = {}

        self._find_skt_lock = asyncio.Lock()

    @abc.abstractmethod
    async def _load_sketch_content(
            self, skt_path: str) -> Union[str, bytes]:  # pragma: no cover
        """
        This is an :func:`abc.abstractmethod`.

        Override this class to customize sketch loading.

        Load the sketch content as string or bytestring.

        .. important::

            If no matched file is found, it should raise a
            :class:`.SketchNotFoundError`.
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def _find_abs_path(
        self, skt_path: str,
            origin_path: Optional[str]=None) -> str:  # pragma: no cover
        """
        This is an :func:`abc.abstractmethod`.

        Override this class to customize sketch discovery.

        Solve the absolute path(starting with :code:`/`) of the sketch from
        the skt_path based on the origin_path(if applicable).

        .. important::

            If no matched file is found, it should raise a
            :class:`.SketchNotFoundError`.
        """
        raise NotImplementedError

    async def _find(
        self, skt_path: str,
            origin_path: Optional[str]=None) -> "sketch.Sketch":
        async with self._find_skt_lock:  # Find one sketch at a time.
            if skt_path in self._skt_cache.keys():
                # Try to read from the cache.
                return self._skt_cache[skt_path]

            # Resolve the path.
            abs_skt_path = await self._find_abs_path(
                skt_path, origin_path=origin_path)

            skt_content = await self._load_sketch_content(abs_skt_path)

            skt = sketch.Sketch(
                skt_content, path=skt_path, skt_ctx=self._ctx,
                finder=self)

            if self._ctx.cache_sketches:
                self._skt_cache[skt_path] = skt

            return skt

    async def find(self, skt_path: str) -> "sketch.Sketch":
        """
        Find the sketch corresponding to the given :code:`skt_path` and
        initialize them with the given :code:`skt_ctx`.

        .. warning::

            If no sketch is matched, this method will raise a
            :class:`.SketchNotFoundError`.
        """
        return await self._find(skt_path)


class SketchFinder(BaseSketchFinder):
    """
    An implementation of :class:`.BaseSketchFinder` using
    `aiofiles <https://github.com/Tinche/aiofiles>`_ to
    load sketches from the local file system asynchronously.

    :arg __root_path: The root path of the finder. Use :code:`/` in including
        and inheritance to indicate this folder. This argument must be passed
        positionally and must be the first argument.
    :arg executor: The executor used by :code:`aiofiles` to load files.
        Default: :code:`None` (Create a new executor upon initialization).
    :arg skt_ctx: The :class:`.SketchContext` to be used by the
        :class:`.SketchFinder` and :class:`.Sketch`. Default: :code:`None`
        (Create a new :class:`.SketchContext` upon initialization).

    """
    def __init__(
        self, __root_path: str, *,
        executor: Optional[concurrent.futures.ThreadPoolExecutor]=None,
            skt_ctx: Optional["context.SketchContext"]=None) -> None:
        assert isinstance(__root_path, str)

        self._root_path = os.path.abspath(__root_path)
        if not self._root_path.endswith("/"):
            self._root_path += "/"

        super().__init__(skt_ctx=skt_ctx)

        self._executor = executor or concurrent.futures.ThreadPoolExecutor()

    @property
    def _loop(self) -> asyncio.AbstractEventLoop:
        return self._ctx.loop

    async def _find_abs_path(
        self, skt_path: str,
            origin_path: Optional[str]=None) -> str:
        skt_path = skt_path.replace("\\", "/")
        # Replace Windows Style Path to UNIX Style.

        if origin_path is not None and (not os.path.isabs(skt_path)):
            origin_dir = os.path.join(
                self._root_path, os.path.dirname(origin_path))

        else:
            origin_dir = self._root_path

        if os.path.isabs(skt_path):
            _, skt_path = skt_path.split("/", 1)
            # Take out the root identifier.

        final_skt_path = os.path.abspath(os.path.join(origin_dir, skt_path))
        final_skt_dir = os.path.dirname(final_skt_path)

        if not final_skt_dir.endswith("/"):
            final_skt_dir += "/"

        if not final_skt_path.startswith(self._root_path):
            raise exceptions.SketchNotFoundError(
                "To prevent potential directory traversal attack, "
                "this path is not acceptable.")

        if not os.path.exists(final_skt_path):
            raise exceptions.SketchNotFoundError(
                f"No such file {final_skt_path}.")

        return final_skt_path

    async def _load_sketch_content(self, skt_path: str) -> bytes:
        async with aiofiles.open(
            skt_path, mode="rb", executor=self._executor,
                loop=self._loop) as skt_fp:
            return await skt_fp.read()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#   Copyright 2017 Kaede Hoshikawa
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

__all__ = ["__version__"]

_tag_version = (0, 1, 0)

_dev = None

_version_fragments = [str(i) for i in _tag_version[:3]]

if _dev is not None:
    _version_fragments.append(f"dev{_dev}")

__version__ = ".".join(_version_fragments)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#   Copyright 2017 Kaede Hoshikawa
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from ._version import *
from . import _version

from .context import *
from . import context

from .exceptions import *
from . import exceptions

from .finders import *
from . import finders

from .runtime import *
from . import runtime

from .sketch import *
from . import sketch

__all__ = _version.__all__ + context.__all__ + exceptions.__all__ + \
    finders.__all__ + runtime.__all__ + sketch.__all__



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#   Copyright 2017 Kaede Hoshikawa
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from setuptools import setup, find_packages

import sys

if not sys.version_info[:3] >= (3, 6, 0):
    raise RuntimeError("Sketchbook requires Python 3.6.0 or higher.")

else:
    try:
        import _modify_version

    except ImportError:
        pass

    else:
        _modify_version.modify("sketchbook")

    import _load_version


setup_requires = ["setuptools", "pytest-runner>=2.11.1,<3"]

install_requires = ["aiofiles>=0.3.1,<1"]
install_requires.extend(setup_requires)

tests_require = ["pytest>=3.0.7,<4"]
tests_require.extend(install_requires)

dev_requires = [
    "sphinx>=1.5.5,<2",
    "sphinxcontrib-versioning>=2.2.1,<3",
    "sphinxcontrib-asyncio>=0.2.0,<1",
    "sphinx_rtd_theme>=0.2.4,<3",
    "mypy>=0.501,<1",
    "pycodestyle>=2.3.1,<3"]
dev_requires.extend(tests_require)

if __name__ == "__main__":
    setup(
        name="sketchbook",
        version=_load_version.load("sketchbook"),
        author="Kaede Hoshikawa",
        author_email="futursolo@icloud.com",
        url="https://github.com/futursolo/sketchbook",
        license="Apache License 2.0",
        description="An ordered, one-to-many mapping.",
        long_description=open("README.rst", "r").read(),
        packages=find_packages(),
        include_package_data=True,
        setup_requires=setup_requires,
        install_requires=install_requires,
        tests_require=tests_require,
        extras_require={
            "test": tests_require,
            "dev": dev_requires,
        },
        zip_safe=False,
        classifiers=[
            "Operating System :: MacOS",
            "Operating System :: MacOS :: MacOS X",
            "Operating System :: Microsoft",
            "Operating System :: Microsoft :: Windows",
            "Operating System :: POSIX",
            "Operating System :: POSIX :: Linux",
            "Operating System :: Unix",
            "Programming Language :: Python",
            "Programming Language :: Python :: 3 :: Only",
            "Programming Language :: Python :: Implementation :: CPython"
        ]
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#   Copyright 2017 Kaede Hoshikawa
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import Dict, Any, Optional, Awaitable, Callable, Type, Union, Tuple

from . import sketch
from . import exceptions
from . import finders
from . import context

import abc

__all__ = ["BlockStorage", "BlockRuntime", "SketchRuntime"]


class BlockStorage:
    """
    A read-only, mapping-like object for :class:`.SketchRuntime` to access
    blocks.

    This class can be accessd using :code:`self.blocks` inside a sketch.

    Example:

    .. code-block:: text

        <% block a %>
            233
        <% end %>
        <%# self object refers to sketchbook.SketchRuntime %>

        <%r= await self.blocks.a() %>    <%# outputs 233 %>
        <%r= await self.blocks["a"]() %> <%# outputs 233 %>
        <%r= await self.blocks.b() %>    <%# raises AttributeError %>
        <%r= await self.blocks["b"]() %> <%# raises KeyError %>

    """
    def __init__(self, skt_rt: "SketchRuntime") -> None:
        self._skt_rt: SketchRuntime
        self._blocks: Dict[str, Type[BlockRuntime]]
        self.__dict__["_skt_rt"] = skt_rt
        self.__dict__["_blocks"] = {}

        self._blocks.update(  # type: ignore
            **self._skt_rt._BLOCK_RUNTIMES)

    def __getitem__(
        self, name: Union[str, Tuple[str, bool]]) -> Callable[
            ..., Awaitable[str]]:
        if isinstance(name, tuple):
            block_name, defined_here = name

        else:
            block_name = name
            defined_here = False

        if block_name not in self._blocks.keys():
            raise KeyError(f"Unknown Block Name {block_name}.")

        SelectedBlockRuntime = self._blocks[block_name]

        async def wrapper() -> str:
            block_rt = SelectedBlockRuntime(
                self._skt_rt, _defined_here=defined_here)

            await block_rt._draw()

            return block_rt._block_result

        return wrapper

    def __setitem__(self, name: str, value: Any) -> None:  # pragma: no cover
        raise NotImplementedError

    def __getattr__(self, name: str) -> Callable[..., Awaitable[str]]:
        try:
            return self[name]

        except KeyError as e:
            raise AttributeError from e

    __setattr__ = __setitem__

    def _update(self, child_store: "BlockStorage") -> None:
        for k, v in child_store._blocks.items():
            if k not in self._blocks.keys():
                continue

            self._blocks[k] = v


class _AbstractRuntime(abc.ABC):
    @property
    @abc.abstractmethod
    def _finder(self) -> "finders.BaseSketchFinder":  # pragma: no cover
        raise NotImplementedError

    @abc.abstractmethod
    def _get_globals(self) -> Dict[str, Any]:  # pragma: no cover
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def _ctx(self) -> "context.SketchContext":  # pragma: no cover
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def blocks(self) -> BlockStorage:  # pragma: no cover
        """
        Return an object to access blocks.
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def body(self) -> str:  # pragma: no cover
        """
        Return the body from the child.
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def parent(self) -> "SketchRuntime":  # pragma: no cover
        """
        Return the runtime of the parent(if any).
        """
        raise NotImplementedError

    @abc.abstractmethod
    def write(
        self, __content: Any,
            escape: str="default") -> None:  # pragma: no cover
        raise NotImplementedError

    @abc.abstractmethod
    async def _inherit_sketch(self) -> None:  # pragma: no cover
        raise NotImplementedError

    @abc.abstractmethod
    async def _add_parent(self, path: str) -> None:  # pragma: no cover
        raise NotImplementedError

    @abc.abstractmethod
    async def _include_sketch(self, path: str) -> str:  # pragma: no cover
        raise NotImplementedError

    @abc.abstractmethod
    async def _draw(self) -> None:  # pragma: no cover
        raise NotImplementedError


class BlockRuntime(_AbstractRuntime):
    """
    The runtime for a block. It shares the same methods and properties with the
    :class:`.SketchRuntime`.

    .. important::

        The content inside a block has a different local bound than the
        other parts of the sketch.
    """
    def __init__(
            self, __skt_rt: "SketchRuntime", _defined_here: bool) -> None:
        self._skt_rt = __skt_rt
        self._defined_here = _defined_here

        self.__skt_result__ = ""

        self._finished = False

    @property
    def _skt(self) -> "sketch.Sketch":
        return self._skt_rt._skt

    @property
    def _finder(self) -> "finders.BaseSketchFinder":
        if self._skt._finder is None:
            raise exceptions.SketchDrawingError(
                "The finder is not set. "
                "The inheritance and including is not available.")
        return self._skt._finder

    def _get_globals(self) -> Dict[str, Any]:
        return self._skt_rt._get_globals()

    @property
    def _ctx(self) -> "context.SketchContext":
        return self._skt_rt._ctx

    @property
    def _block_result(self) -> str:
        if not self._finished:
            raise exceptions.SketchDrawingError(
                "Drawing has not been finished yet.")

        return self.__skt_result__

    @property
    def blocks(self) -> BlockStorage:
        return self._skt_rt.blocks

    def write(self, __content: Any, escape: str="default") -> None:
        if self._finished:
            raise exceptions.SketchDrawingError(
                "Drawing has been finished.")

        self.__skt_result__ += self._ctx.escape_fns[escape](__content)

    @property
    def body(self) -> str:
        return self._skt_rt.body

    @property
    def parent(self) -> "SketchRuntime":
        return self._skt_rt.parent

    async def _inherit_sketch(self) -> None:  # pragma: no cover
        raise exceptions.SketchDrawingError(
            "Cannot inherit sketch(es) inside the block.")

    async def _add_parent(self, path: str) -> None:  # pragma: no cover
        raise exceptions.SketchDrawingError(
            "Cannot Set Inheritance inside the block.")

    async def _include_sketch(self, path: str) -> str:
        return await self._skt_rt._include_sketch(path)

    async def _draw(self) -> None:
        if self._finished:
            raise exceptions.SketchDrawingError(
                "Drawing has already been finished.")

        if self._defined_here and self._skt_rt._parent is not None:
            self._finished = True

            return

        await self._draw_block()

        self._finished = True

    @abc.abstractmethod
    async def _draw_block(self) -> None:  # pragma: no cover
        raise NotImplementedError


class SketchRuntime(_AbstractRuntime):
    """
    The Sketch Runtime -- the :code:`self` inside sketches.

    The runtime provides a series of methods and properties and to help
    developers when drawing the sketch.
    """
    _BLOCK_RUNTIMES: Dict[str, Type[BlockRuntime]] = {}

    def __init__(
        self, skt: "sketch.Sketch",
            skt_globals: Dict[str, Any]) -> None:
        self._skt = skt
        self._skt_globals = skt_globals

        self.__skt_result__ = ""

        self._body: Optional[str] = None
        self._parent: Optional[SketchRuntime] = None

        self._block_store = BlockStorage(self)

        self._finished = False

    @property
    def _finder(self) -> "finders.BaseSketchFinder":
        if self._skt._finder is None:
            raise exceptions.SketchDrawingError(
                "The finder is not set. "
                "The inheritance and including is not available.")
        return self._skt._finder

    def _get_globals(self) -> Dict[str, Any]:
        new_globals = {}

        new_globals.update(self._skt_globals)

        if "_SktCurrentRuntime" in new_globals.keys():
            del new_globals["_SktCurrentRuntime"]

        return new_globals

    @property
    def _skt_result(self) -> str:
        if not self._finished:
            raise exceptions.SketchDrawingError(
                "Drawing has not been finished yet.")

        return self.__skt_result__

    @property
    def _ctx(self) -> "context.SketchContext":
        return self._skt._ctx

    @property
    def blocks(self) -> BlockStorage:
        """
        Return an :class:`.BlockStorage` to help developers to access blocks.
        """
        return self._block_store

    def write(self, __content: Any, escape: str="default") -> None:
        """
        Write the content to the buffer.

        :arg __content: The content to write. This must be str, must be passed
            as a positional argument, and must be the first argument of the
            function call.
        :arg escape: The name of the escape function to use.
            For more information, see: :class:`.SketchContext`.
            Default: :code:`default`.
        """
        if self._finished:
            raise exceptions.SketchDrawingError(
                "Drawing has been finished.")

        self.__skt_result__ += self._ctx.escape_fns[escape](__content)

    @property
    def body(self) -> str:
        """
        The body of the child sketch(if any).

        .. warning::

            If the current sketch is not the parent of another sketch,
            it will raise an :class:`AttributeError`.
        """
        if self._body is None:
            raise AttributeError("Inheritance is not enabled.")

        return self._body

    @property
    def parent(self) -> "SketchRuntime":
        """
        The runtime of the parent sketch(if any).

        .. warning::

            If the current sketch is not the child of another sketch,
            it will raise an :class:`AttributeError`.
        """
        if self._parent is None:
            raise AttributeError("Parent is not set.")

        return self._parent

    def _update_blocks(self, child_store: BlockStorage) -> None:
        self._block_store._update(child_store)

    def _update_body(self, body: str) -> None:
        assert self._body is None, "There's already a child body."
        self._body = body

    async def _inherit_sketch(self) -> None:
        if self._parent is None:
            return

        self._parent._update_body(self.__skt_result__)
        self._parent._update_blocks(self._block_store)

        await self._parent._draw()

        self.__skt_result__ = self._parent._skt_result

    async def _add_parent(self, path: str) -> None:
        assert self._parent is None, \
            "A sketch can only set the inheritance once."

        parent_skt = await self._finder._find(
            path, origin_path=self._skt._path)

        self._parent = parent_skt._get_runtime(
            skt_globals=self._get_globals())

    async def _include_sketch(self, path: str) -> str:
        skt = await self._finder._find(path, origin_path=self._skt._path)

        skt_rt = skt._get_runtime(self._get_globals())

        await skt_rt._draw()
        return skt_rt._skt_result

    async def _draw(self) -> None:
        if self._finished:
            raise exceptions.SketchDrawingError(
                "Drawing has already been finished.")

        await self._draw_body()

        await self._inherit_sketch()
        self._finished = True

    @abc.abstractmethod
    async def _draw_body(self) -> None:  # pragma: no cover
        raise NotImplementedError



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#   Copyright 2017 Kaede Hoshikawa
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import List, Optional, Sequence, Type, Dict

from . import printer
from . import exceptions
from . import sketch

import abc
import re

__all__ = [
    "Statement", "Root", "Block", "BaseOutput", "builtin_stmt_classes",
    "IndentMixIn", "AppendMixIn", "UnindentMixIn"]

_VALID_FN_NAME_RE = re.compile(r"^[a-zA-Z]([a-zA-Z0-9\_]+)?$")


def _is_valid_fn_name(maybe_fn_name: str) -> bool:
    """
    Check if this is a valid function name.
    """
    return (re.fullmatch(_VALID_FN_NAME_RE, maybe_fn_name) is not None)


class IndentMixIn(abc.ABC):
    @abc.abstractmethod
    def append_stmt(self, stmt: "AppendMixIn") -> None:  # pragma: no cover
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def line_no(self) -> int:  # pragma: no cover
        raise NotImplementedError


class AppendMixIn(abc.ABC):
    @property
    @abc.abstractmethod
    def line_no(self) -> int:  # pragma: no cover
        raise NotImplementedError

    @abc.abstractmethod
    def print_code(
        self, py_printer: printer.PythonPrinter) -> \
            None:  # pragma: no cover
        raise NotImplementedError


class UnindentMixIn(abc.ABC):
    pass


class Statement(abc.ABC):
    @classmethod
    @abc.abstractmethod
    def try_match(
        Cls, stmt_str: str, skt: sketch.Sketch,
            line_no: int) -> Optional["Statement"]:  # pragma: no cover
        raise NotImplementedError


class Root(Statement, AppendMixIn):
    def __init__(self, skt: sketch.Sketch) -> None:
        self._skt = skt

        self._stmts: List[AppendMixIn] = []
        self._block_stmts: Dict[str, Block] = {}

    @property
    def line_no(self) -> int:
        return 0

    def append_stmt(self, stmt: AppendMixIn) -> None:
        self._stmts.append(stmt)

    def append_block(self, block_stmt: "Block") -> None:
        if block_stmt.block_name in self._block_stmts.keys():
            raise exceptions.BlockNameConflictError(
                f"The name of the block at {block_stmt.line_no}"
                "is conflict with the block at "
                f"{self._block_stmts[block_stmt.block_name].line_no} "
                f"in file {self._skt._path}. You cannot define two block with "
                "the same name in the one file.")

        self._block_stmts[block_stmt.block_name] = block_stmt

    @classmethod
    def try_match(
        Cls, stmt_str: str, skt: sketch.Sketch,
            line_no: int) -> Optional["Statement"]:  # pragma: no cover
        raise NotImplementedError("This does not apply to Root.")

    def print_code(self, py_printer: printer.PythonPrinter) -> None:
        py_printer.writeline("import sketchbook")
        py_printer.writeline("_SKT_BLOCK_RUNTIMES = {}")

        for block_stmt in self._block_stmts.values():
            block_stmt.print_block_code(py_printer)

        py_printer.writeline(
            "class _SktCurrentRuntime(sketchbook.SketchRuntime):", self)
        with py_printer.indent_block():
            py_printer.writeline(
                "_BLOCK_RUNTIMES = _SKT_BLOCK_RUNTIMES", self)

            py_printer.writeline(
                "async def _draw_body(self) -> None:", self)
            with py_printer.indent_block():
                for stmt in self._stmts:
                    stmt.print_code(py_printer)


class Block(Statement, IndentMixIn, AppendMixIn):
    def __init__(
            self, block_name: str, skt: sketch.Sketch, line_no: int) -> None:
        self._block_name = block_name
        self._skt = skt
        self._line_no = line_no

        self._stmts: List[AppendMixIn] = []

    def append_stmt(self, stmt: AppendMixIn) -> None:
        self._stmts.append(stmt)

    @property
    def block_name(self) -> str:
        return self._block_name

    @property
    def line_no(self) -> int:
        return self._line_no

    @classmethod
    def try_match(
        Cls, stmt_str: str, skt: sketch.Sketch,
            line_no: int) -> Optional["Statement"]:
        splitted_stmt = stmt_str.split(" ", 1)
        if splitted_stmt[0] != "block":
            return None

        if len(splitted_stmt) != 2 or (not splitted_stmt[1].strip()):
            raise exceptions.SketchSyntaxError(
                "Block name cannot be empty.")

        block_name = splitted_stmt[1].strip()

        if not _is_valid_fn_name(block_name):
            raise exceptions.SketchSyntaxError(
                "Invalid Block Statement. Block name expected, "
                f"got: {repr(block_name)}.")

        return Cls(
            block_name=block_name,
            skt=skt, line_no=line_no)

    def print_code(self, py_printer: printer.PythonPrinter) -> None:
        py_printer.writeline(
            f"self.write(await self.blocks[{repr(self.block_name)}, True](), "
            "escape=\"raw\")",
            self)

    def print_block_code(self, py_printer: printer.PythonPrinter) -> None:
        py_printer.writeline(
            "class _SktCurrentBlockRuntime(sketchbook.BlockRuntime):",
            self)
        with py_printer.indent_block():
            py_printer.writeline(
                "async def _draw_block(self) -> None:", self)
            with py_printer.indent_block():
                for stmt in self._stmts:
                    stmt.print_code(py_printer)

        py_printer.writeline(
            f"_SKT_BLOCK_RUNTIMES[{self.block_name!r}] = "
            "_SktCurrentBlockRuntime")


class Plain(Statement, AppendMixIn):
    def __init__(self, plain_str: str) -> None:
        self._plain_str = plain_str

    @property
    def line_no(self) -> int:  # pragma: no cover
        raise NotImplementedError("This does not apply to Plain.")

    @classmethod
    def try_match(
        Cls, stmt_str: str, skt: sketch.Sketch,
            line_no: int) -> Optional["Statement"]:  # pragma: no cover
        raise NotImplementedError("This does not apply to Plain.")

    def print_code(self, py_printer: printer.PythonPrinter) -> None:
        py_printer.writeline(
            f"self.write({repr(self._plain_str)}, escape=\"raw\")")


class BaseOutput(Statement, AppendMixIn):
    _filter_fn_names: List[str] = []

    def __init__(
        self, output_filter: str, output_exp: str,
            skt: sketch.Sketch, line_no: int) -> None:
        self._output_filter = output_filter
        self._output_exp = output_exp
        self._skt = skt
        self._line_no = line_no

    @property
    def line_no(self) -> int:
        return self._line_no

    @classmethod
    def try_match(
            Cls, stmt_str: str, skt: sketch.Sketch,
            line_no: int) -> Optional["Statement"]:
        splitted_stmt = stmt_str.split(" ", 1)

        stmt_keyword = splitted_stmt[0]

        if not stmt_keyword.endswith("="):
            return None

        stmt_output_filter = stmt_keyword[:-1] or "default"

        if stmt_output_filter not in Cls._filter_fn_names:
            return None

        if len(splitted_stmt) != 2:
            raise exceptions.SketchSyntaxError(
                ("The expression to be output is empty "
                 f"in file {skt._path} at line {line_no}."))

        stmt_output_exp = splitted_stmt[1].strip()

        if not stmt_output_exp:
            raise exceptions.SketchSyntaxError(
                ("The expression to be output is empty "
                 f"in file {skt._path} at line {line_no}."))

        return Cls(
            output_filter=stmt_output_filter,
            output_exp=stmt_output_exp,
            skt=skt, line_no=line_no)

    def print_code(self, py_printer: printer.PythonPrinter) -> None:
        py_printer.writeline(
            f"self.write({self._output_exp}, "
            f"escape={self._output_filter!r})")


class _Include(Statement, AppendMixIn):
    def __init__(
            self, target_path: str, skt: sketch.Sketch, line_no: int) -> None:
        self._target_path = target_path
        self._skt = skt
        self._line_no = line_no

    @property
    def line_no(self) -> int:
        return self._line_no

    @classmethod
    def try_match(
            Cls, stmt_str: str, skt: sketch.Sketch,
            line_no: int) -> Optional["Statement"]:

        splitted_stmt = stmt_str.split(" ", 1)
        if splitted_stmt[0] != "include":
            return None

        if len(splitted_stmt) < 2:
            raise exceptions.SketchSyntaxError(
                f"In file {skt._path} at line {line_no}, "
                "invalid syntax, you must provide the path to be included.")

        return Cls(
            target_path=splitted_stmt[1],
            skt=skt, line_no=line_no)

    def print_code(self, py_printer: printer.PythonPrinter) -> None:
        py_printer.writeline(
            "self.write(await "
            f"self._include_sketch({self._target_path}), escape=\"raw\")",
            self)


class _Inherit(Statement, AppendMixIn):
    def __init__(
            self, target_path: str, skt: sketch.Sketch, line_no: int) -> None:
        self._target_path = target_path
        self._skt = skt
        self._line_no = line_no

    @property
    def line_no(self) -> int:
        return self._line_no

    @classmethod
    def try_match(
        Cls, stmt_str: str, skt: sketch.Sketch,
            line_no: int) -> Optional["Statement"]:
        splitted_stmt = stmt_str.split(" ", 1)
        if splitted_stmt[0] != "inherit":
            return None

        if len(splitted_stmt) < 2:
            raise exceptions.SketchSyntaxError(
                f"In file {skt._path} at line {line_no}, "
                "invalid syntax, you must provide the path to be inherited.")

        return Cls(
            target_path=splitted_stmt[1],
            skt=skt, line_no=line_no)

    def print_code(self, py_printer: printer.PythonPrinter) -> None:
        py_printer.writeline(
            f"await self._add_parent({self._target_path})", self)


class _Indent(Statement, IndentMixIn, AppendMixIn):
    def __init__(
            self, stmt_str: str, skt: sketch.Sketch, line_no: int) -> None:
        self._stmt_str = stmt_str
        self._skt = skt
        self._line_no = line_no

        self._stmts: List[AppendMixIn] = []

    def append_stmt(self, stmt: AppendMixIn) -> None:
        self._stmts.append(stmt)

    @property
    def line_no(self) -> int:
        return self._line_no

    @classmethod
    def try_match(
        Cls, stmt_str: str, skt: sketch.Sketch,
            line_no: int) -> Optional["Statement"]:
        if stmt_str.split(" ", 1)[0] not in (
                "if", "with", "for", "while", "try", "async"):
            return None

        return Cls(
            stmt_str=stmt_str.strip(),
            skt=skt, line_no=line_no)

    def print_code(self, py_printer: printer.PythonPrinter) -> None:
        py_printer.writeline(f"{self._stmt_str}:", self)

        with py_printer.indent_block():
            for stmt in self._stmts:
                stmt.print_code(py_printer)

            py_printer.writeline("pass", self)


class _Unindent(Statement, UnindentMixIn):
    @classmethod
    def try_match(
        Cls, stmt_str: str, skt: sketch.Sketch,
            line_no: int) -> Optional["Statement"]:
        if stmt_str.split(" ", 1)[0] != "end":
            return None

        return Cls()


class _HalfIndent(Statement, IndentMixIn, AppendMixIn, UnindentMixIn):
    def __init__(
            self, stmt_str: str, skt: sketch.Sketch, line_no: int) -> None:
        self._stmt_str = stmt_str
        self._skt = skt
        self._line_no = line_no

        self._stmts: List[AppendMixIn] = []

    def append_stmt(self, stmt: AppendMixIn) -> None:
        self._stmts.append(stmt)

    @property
    def line_no(self) -> int:
        return self._line_no

    @classmethod
    def try_match(
        Cls, stmt_str: str, skt: sketch.Sketch,
            line_no: int) -> Optional["Statement"]:
        if stmt_str.split(" ", 1)[0] not in (
                "else", "elif", "except", "finally"):
            return None

        return Cls(
            stmt_str=stmt_str.strip(),
            skt=skt, line_no=line_no)

    def print_code(self, py_printer: printer.PythonPrinter) -> None:
        py_printer.writeline(f"{self._stmt_str}:", self)

        with py_printer.indent_block():
            for stmt in self._stmts:
                stmt.print_code(py_printer)

            py_printer.writeline("pass", self)


class _Inline(Statement, AppendMixIn):
    def __init__(
            self, stmt_str: str, skt: sketch.Sketch, line_no: int) -> None:
        self._stmt_str = stmt_str
        self._skt = skt
        self._line_no = line_no

    @property
    def line_no(self) -> int:
        return self._line_no

    @classmethod
    def try_match(
        Cls, stmt_str: str, skt: sketch.Sketch,
            line_no: int) -> Optional["Statement"]:
        if stmt_str.split(" ", 1)[0] not in (
            "break", "continue", "import", "raise", "from",
                "nonlocal", "global", "assert"):
            return None

        return Cls(
            stmt_str=stmt_str.strip(),
            skt=skt, line_no=line_no)

    def print_code(self, py_printer: printer.PythonPrinter) -> None:
        py_printer.writeline(self._stmt_str, self)


class _Comment(Statement, AppendMixIn):
    def __init__(
            self, cmnt_str: str, skt: sketch.Sketch, line_no: int) -> None:
        self._cmnt_str = cmnt_str
        self._skt = skt
        self._line_no = line_no

    @property
    def line_no(self) -> int:
        return self._line_no

    @classmethod
    def try_match(
        Cls, stmt_str: str, skt: sketch.Sketch,
            line_no: int) -> Optional["Statement"]:
        if not stmt_str.startswith("#"):
            return None

        return Cls(
            cmnt_str=stmt_str[1:].strip(),
            skt=skt, line_no=line_no)

    def print_code(self, py_printer: printer.PythonPrinter) -> None:
        for cmnt_line in self._cmnt_str.splitlines():
            py_printer.writeline(f"# {cmnt_line}", self)


class _Assign(Statement, AppendMixIn):
    def __init__(
        self, target_lst: str, exp: str, skt: sketch.Sketch,
            line_no: int) -> None:
        self._target_lst = target_lst
        self._exp = exp

        self._skt = skt
        self._line_no = line_no

    @property
    def line_no(self) -> int:
        return self._line_no

    @classmethod
    def try_match(
        Cls, stmt_str: str, skt: sketch.Sketch,
            line_no: int) -> Optional["Statement"]:
        stmt_str = stmt_str.strip()
        if not stmt_str.startswith("let "):
            return None

        assign_stmt = stmt_str[4:]

        try:
            target_lst, exp = assign_stmt.split("=")

        except (ValueError, TypeError):
            raise exceptions.SketchSyntaxError(
                f"Invaild assignment statment at line {line_no} "
                f"in file {skt._path}.")

        return Cls(
            target_lst=target_lst.strip(),
            exp=exp.strip(),
            skt=skt, line_no=line_no)

    def print_code(self, py_printer: printer.PythonPrinter) -> None:
        py_printer.writeline(f"{self._target_lst} = {self._exp}", self)


builtin_stmt_classes: Sequence[Type[Statement]] = [
    Block, _Include, _Inherit, _Indent, _Unindent, _HalfIndent, _Inline,
    _Comment, _Assign]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#   Copyright 2017 Kaede Hoshikawa
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import Any, Optional
from types import CodeType

from . import sketch
from . import statements

__all__ = ["PythonPrinter"]


class PythonPrinter:
    """
    Print Python code with indentation gracefully.
    """
    def __init__(
        self, path: str="<string>", end: str="\n",
            indent_mark: str=" " * 4) -> None:
        self._path = path
        self._indent_num = 0
        self._committed_code = ""

        self._end = end
        self._indent_mark = indent_mark

        self._finished = False

    def writeline(
        self, line: str,
            stmt: Optional["statements.AppendMixIn"]=None) -> None:
        """
        Write a line with indent.
        """
        assert not self._finished, "Code Generation has already been finished."

        if stmt:
            line += f"  # in file {self._path} at line {stmt.line_no}."

        final_line = self._indent_mark * self._indent_num + line + self._end
        self._committed_code += final_line

    def indent_block(self) -> "PythonPrinter":
        """
        Indent the code with `with` statement.

        Example::

              printer.writeline("def a():")
              with printer.indent_block():
                  printer.writeline("return \"Text from function a.\"")

              printer.writeline("a()")

        """
        assert not self._finished, "Code Generation has already been finished."
        return self

    def _inc_indent_num(self) -> None:
        assert not self._finished, "Code Generation has already been finished."
        self._indent_num += 1

    def _dec_indent_num(self) -> None:
        assert not self._finished, "Code Generation has already been finished."
        self._indent_num -= 1

    def __enter__(self) -> None:
        self._inc_indent_num()

    def __exit__(self, *exc: Any) -> None:
        self._dec_indent_num()

    @property
    def finished(self) -> bool:  # pragma: no cover
        return self._finished

    @property
    def plain_code(self) -> str:
        """
        Return the plain, printed code.
        """
        self._finished = True
        return self._committed_code

    @property
    def compiled_code(self) -> CodeType:
        """
        Return the compiled code.
        """
        if not hasattr(self, "_compiled_code"):
            self._compiled_code = compile(
                self.plain_code, self._path, "exec", dont_inherit=True)

        return self._compiled_code

    @classmethod
    def print_sketch(Cls, skt: "sketch.Sketch") -> CodeType:
        py_printer = Cls(path=skt._path)
        skt._root.print_code(py_printer)
        return py_printer.compiled_code



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#   Copyright 2017 Kaede Hoshikawa
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import importlib
import os


def load(project_folder: str) -> str:
    _version_spec = importlib.util.spec_from_file_location(
        f"{project_folder}._version",
        os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            f"{project_folder}/_version.py"))

    _version = importlib.util.module_from_spec(_version_spec)
    _version_spec.loader.exec_module(_version)

    return _version.__version__



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#   Copyright 2017 Kaede Hoshikawa
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

__all__ = [
    "SketchbookException", "SketchNotFoundError", "SketchSyntaxError",
    "UnknownStatementError", "BlockNameConflictError", "SketchDrawingError"]


class SketchbookException(Exception):
    """
    Base class of exceptions from Sketchbook.
    """
    pass


class SketchNotFoundError(FileNotFoundError, SketchbookException):
    """
    Error when trying to load a sketch but the finder cannot find it.
    """
    pass


class SketchSyntaxError(SyntaxError, SketchbookException):
    """
    Syntax error in the current sketch.
    """
    pass


class UnknownStatementError(SketchSyntaxError):
    """
    The statement string is not matched by any known Statement Classes.
    """
    pass


class BlockNameConflictError(SketchbookException):
    """
    There's more than one block with the same name in one sketch.
    """
    pass


class SketchDrawingError(SketchbookException):
    """
    Error when drawing the sketch.
    """
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#   Copyright 2017 Kaede Hoshikawa
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import Optional, List

from . import statements
from . import sketch
from . import exceptions

import io

__all__ = ["SketchParser"]


class _ReadFinished(Exception):
    pass


class SketchParser:
    """
    The one-time, non-reusable sketch parser.
    """
    def __init__(self, __skt: "sketch.Sketch") -> None:
        self._skt = __skt
        self._ctx = self._skt._ctx

        self._io = io.StringIO(self._skt._content)

        self._current_line_no = 0
        self._current_line = ""

        self._root = statements.Root(skt=self._skt)
        self._indents: List[statements.IndentMixIn] = []

        self._parse()

    def _move_to_next_line(self) -> None:
        assert not self._current_line, \
            ("Parsing of the last line is not completed "
             f"in file {self._skt._path} at line {self._current_line_no}.")

        new_line = self._io.readline()

        if not new_line:
            raise _ReadFinished

        self._current_line_no += 1
        self._current_line = new_line

    @property
    def root(self) -> "statements.Root":
        return self._root

    def _find_next_begin_mark(self) -> int:
        start_pos = 0

        while True:
            pos = self._current_line.find("<%", start_pos)

            if pos == -1:
                return -1

            elif self._current_line[pos + 2:].startswith("%"):
                start_pos = pos + 2
                end_pos = start_pos + 1

                self._current_line = (
                    self._current_line[:start_pos] +
                    self._current_line[end_pos:])

                continue

            else:
                return pos

    def _find_next_end_mark(self) -> int:
        start_pos = 0

        while True:
            pos = self._current_line.find("%>", start_pos)

            if pos == -1:
                return -1

            elif pos == 0:
                return 0

            elif self._current_line[:pos].endswith("%"):
                start_pos = pos - 1
                end_pos = pos

                self._current_line = (
                    self._current_line[:start_pos] +
                    self._current_line[end_pos:])

                start_pos = pos + 1

                continue

            else:
                return pos

    def _parse_stmt(
            self, stmt_str: str, line_no: int) -> "statements.Statement":
        stmt_str = stmt_str.strip()

        if stmt_str.endswith(":"):
            raise exceptions.SketchSyntaxError(
                    "A Statement Cannot be Finished with `:`, "
                    "use `<% end %>` instead. For detailed information, "
                    "please see the documation.")

        for StmtCls in self._ctx.stmt_classes:
            maybe_stmt = StmtCls.try_match(
                stmt_str, skt=self._skt, line_no=line_no)

            if maybe_stmt is None:
                continue

            return maybe_stmt

        else:
            raise exceptions.UnknownStatementError(
                f"Unknown Statement {repr(stmt_str)} "
                f"in file {self._skt._path} at line {line_no}.")

    def _find_next_stmt(self) -> "statements.Statement":
        stmt_chunks: List[str] = []
        begin_mark_line_no = -1

        while True:
            if not self._current_line:
                try:
                    self._move_to_next_line()

                except _ReadFinished:
                    if begin_mark_line_no != -1:
                        raise exceptions.SketchSyntaxError(
                            ("Cannot find end mark for begin mark "
                             f"in file {self._skt._path} "
                             f"at line {begin_mark_line_no}."))

                    elif stmt_chunks:
                        return statements.Plain("".join(stmt_chunks))

                    else:
                        raise

            if begin_mark_line_no == -1:
                begin_mark_pos = self._find_next_begin_mark()

                if begin_mark_pos == -1:
                    stmt_chunks.append(self._current_line)
                    self._current_line = ""

                    continue

                elif begin_mark_pos > 0:
                    stmt_chunks.append(self._current_line[:begin_mark_pos])
                    self._current_line = self._current_line[begin_mark_pos:]

                    return statements.Plain("".join(stmt_chunks))

                elif stmt_chunks:
                    return statements.Plain("".join(stmt_chunks))

                begin_mark_line_no = self._current_line_no
                self._current_line = self._current_line[2:]

            end_mark_pos = self._find_next_end_mark()

            if end_mark_pos == -1:
                stmt_chunks.append(self._current_line)
                self._current_line = ""

                continue

            stmt_chunks.append(self._current_line[:end_mark_pos])
            self._current_line = self._current_line[end_mark_pos + 2:]

            return self._parse_stmt("".join(stmt_chunks), begin_mark_line_no)

    def _append_to_current(self, stmt: "statements.AppendMixIn") -> None:
        if self._indents:
            self._indents[-1].append_stmt(stmt)

        else:
            self._root.append_stmt(stmt)

    def _unindent_current(self) -> None:
        if self._indents:
            self._indents.pop()

        else:
            raise exceptions.SketchSyntaxError(
                "Redundant Unindent Statement "
                f"in file {self._skt._path} at line {self._current_line_no}.")

    def _parse(self) -> None:
        while True:
            try:
                stmt = self._find_next_stmt()

            except _ReadFinished:
                break

            if isinstance(stmt, statements.UnindentMixIn):
                self._unindent_current()

            if isinstance(stmt, statements.AppendMixIn):
                self._append_to_current(stmt)

            if isinstance(stmt, statements.IndentMixIn):
                self._indents.append(stmt)

            if isinstance(stmt, statements.Block):
                self._root.append_block(stmt)

        if self._indents:
            raise exceptions.SketchSyntaxError(
                "Unindented Indent Statement "
                f"in file {self._skt._path} "
                f"at line {self._indents[-1].line_no}.")

    @classmethod
    def parse_sketch(Cls, skt: "sketch.Sketch") -> "statements.Root":
        skt_parser = Cls(skt)
        return skt_parser.root
